import logging
import os
import time
from random import random

from . import Monitor


class LOGMonitor(Monitor):
    created = False

    def __init__(self, name: str):
        super().__init__(name)

    def create(self):
        logging.basicConfig(
            format=f"[{os.getpid()}]->%(asctime)s %(levelname)-4s:%(message)s",
            filename=f"{self.name}.log",
            encoding="utf-8",
            level=logging.INFO,
            datefmt="%Y-%m-%d %H:%M:%S",
            force=True,
        )
        self.created = True

    def track_put(self):
        if not self.created:
            self.create()
        self.put_counter += 1

        logging.info(f"put counter: {self.put_counter}")

    def track_get(self):
        if not self.created:
            self.create()
        self.get_counter += 1
        logging.info(f"get counter: {self.get_counter}")

    def time_in_queue(self, tnq: int):
        if not self.created:
            self.create()
        logging.info(f"time-in-queue {tnq/1000000} millisecond")

    def track_time_in_pool(self, nano_seconds: int):
        if not self.created:
            self.create()
        logging.info(f"time-in-pool {nano_seconds/1000000} millisecond")



import time
from random import random

from opencensus.ext.stackdriver import stats_exporter
from opencensus.stats import aggregation, measure, stats, view

from . import Monitor


class GCPMonitor(Monitor):
    created = False

    def __init__(self, name: str):
        super().__init__(name)

    def _create_put_measure(self):
        self.m_put = measure.MeasureInt(f"mnqueues.put", "number of puts", "1")
        self.v_put = view.View(
            f"mnqueues.{self.name}.number_queue_put",
            "number of put() to queues",
            [],
            self.m_put,
            aggregation.CountAggregation(),
        )
        stats.stats.view_manager.register_view(self.v_put)

    def _create_get_measure(self):
        self.m_get = measure.MeasureInt("mnqueues.get", "number of gets", "1")
        self.v_get = view.View(
            f"mnqueues.{self.name}.number_queue_get",
            "number of get() to queues",
            [],
            self.m_get,
            aggregation.CountAggregation(),
        )
        stats.stats.view_manager.register_view(self.v_get)

    def _create_tnq_measure(self):
        self.m_tnq = measure.MeasureInt("mnqueues.mnq", "time in queue", "ms")
        self.v_tnq = view.View(
            f"mnqueues.{self.name}.time_in_queue_distribution",
            "The distribution of the queue latencies",
            [],
            self.m_tnq,
            aggregation.DistributionAggregation(
                [10, 50, 100, 500, 1000, 10000]
            ),
        )
        stats.stats.view_manager.register_view(self.v_tnq)

    def _create_time_in_pool_measure(self):
        self.m_tnp = measure.MeasureInt("mnqueues.mnp", "time in pool", "ms")
        self.v_tnp = view.View(
            f"mnqueues.{self.name}.time_in_pool_distribution",
            "The distribution of time in pool latencies",
            [],
            self.m_tnp,
            aggregation.DistributionAggregation(
                [10, 50, 100, 500, 1000, 10000]
            ),
        )
        stats.stats.view_manager.register_view(self.v_tnp)

    def create(self):
        self.created = True

        self._create_put_measure()
        self._create_get_measure()
        self._create_tnq_measure()
        self._create_time_in_pool_measure()

        exporter = stats_exporter.new_stats_exporter()
        stats.stats.view_manager.register_exporter(exporter)

    def track_put(self):
        if not self.created:
            self.create()
        self.put_counter += 1
        mmap = stats.stats.stats_recorder.new_measurement_map()
        mmap.measure_int_put(self.m_put, self.put_counter)
        mmap.record()

    def track_get(self):
        if not self.created:
            self.create()
        self.get_counter += 1
        mmap = stats.stats.stats_recorder.new_measurement_map()
        mmap.measure_int_put(self.m_get, self.get_counter)
        mmap.record()

    def time_in_queue(self, tnq: int):
        if not self.created:
            self.create()
        mmap = stats.stats.stats_recorder.new_measurement_map()
        mmap.measure_int_put(self.m_tnq, tnq / 1000000)
        mmap.record()

    def track_time_in_pool(self, nano_seconds: int):
        if not self.created:
            self.create()
        mmap = stats.stats.stats_recorder.new_measurement_map()
        mmap.measure_int_put(self.m_tnp, nano_seconds / 1000000)
        mmap.record()



import multiprocessing as mp
from typing import Optional

__version__ = "0.0.33"

import copy
import time
from random import randint


class Monitor:
    def __init__(self, name: str):
        self.name = name
        self.put_counter: int = 0
        self.get_counter: int = 0

    def track_put(self):
        pass

    def track_get(self):
        pass

    # tnq is measured in nano-seconds
    def time_in_queue(self, tnq: int):
        pass

    def track_time_in_pool(self, nano_seconds: int):
        pass


class MNQueue:
    def __init__(self, monitor: Optional[Monitor] = None, maxsize=0):
        self.monitor = monitor
        self.queue: mp.Queue = mp.Queue(maxsize)

    def put_w_tnq(self, *args, **kwargs):
        payload = {
            "_signature": "mnqueue_tnq",
            "tnq": time.time_ns(),
        }
        if "obj" in kwargs:
            payload["obj"] = kwargs["obj"]
            kwargs["obj"] = payload
        else:
            payload["obj"] = args[0]
            args = list(args)
            args[0] = payload
            args = tuple(args)

        return self.queue.put(*args, **kwargs)

    def should_send_tnq(self) -> bool:
        return randint(1, 20) == 1  # nosec

    def put(self, *args, **kwargs):
        if self.monitor:
            self.monitor.track_put()
            if self.should_send_tnq():
                return self.put_w_tnq(*args, **kwargs)

        return self.queue.put(*args, **kwargs)

    def extract_tnq(self, queue_content):
        if (
            isinstance(queue_content, dict)
            and queue_content.get("_signature", "") == "mnqueue_tnq"
        ):
            self.monitor.time_in_queue(
                time.time_ns() - int(queue_content["tnq"])
            )
            return queue_content["obj"]
        return queue_content

    def get(self, *args, **kwargs):
        rc = self.queue.get(*args, **kwargs)
        if self.monitor:
            self.monitor.track_get()
            rc = self.extract_tnq(rc)

        return rc

    def __getstate__(self):
        print("MNQueue pickling")
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state
        print("MNQueue unpickling")

    def __getattr__(self, attr):
        if self.queue and attr not in self.__dict__:
            return self.queue.__getattribute__(attr)


class MNPool:
    def __init__(self, *args, monitor=None, **kwargs):
        self.monitor = monitor
        self.pool: mp.Pool = mp.Pool(*args, **kwargs)
        self.tnp: int

    def __getattr__(self, attr):
        if attr not in self.__dict__:
            return self.pool.__getattribute__(attr)

    def __enter__(self):
        if self.monitor:
            self.tnp = time.time_ns()

        return self.pool.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        _exit = self.pool.__exit__(exc_type, exc_val, exc_tb)

        if self.monitor:
            self.monitor.track_time_in_pool(time.time_ns() - self.tnp)

        return _exit
